;-----------------------------------------------------------
;
; TGX.ASM - Graphics in text mode
; Copyright (c) 1993-2003,2004 by Alessandro Scotti
; http://www.ascotti.org/
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;-----------------------------------------------------------
.286P
JUMPS
LOCALS
WARN    PRO
NOWARN  RES
INCLUDE TGX.DEF

PUBLIC  tgxInit
PUBLIC  tgxDone
PUBLIC  tgxShow
PUBLIC  tgxClear
PUBLIC  tgxRefresh
PUBLIC  tgxGetPixel
PUBLIC  tgxSetPixel
PUBLIC  tgxResetPixel
PUBLIC  tgxGrabChar
PUBLIC  tgxDoneWindow
PUBLIC  tgxInitWindow
PUBLIC  tgxWaitRetrace

TGX_TEXT        SEGMENT BYTE PUBLIC 'CODE'
                ASSUME cs:TGX_TEXT, ds:NOTHING, es:NOTHING

; Masks to set/reset pixels
baOrPixelMask   DB      128, 64, 32, 16, 8, 4, 2, 1
baAndPixelMask  DB      127, 191, 223, 239, 247, 251, 253, 254

boDPMI          DB      FALSE           ; DPMI in protected mode flag
boInit          DB      FALSE           ; Initialization flag

wSegData        DW      ?               ; 'DATA' alias for code segment
wSegA000h       DW      0A000h          ; Selector for segment A000h
wSegB800h       DW      0B800h          ;    "      "     "    B800h
wCharHeight     DW      ?               ; Character height in pixels
wScreenWidth    DW      ?               ; Screen columns
wScreenHeight   DW      ?               ; Screen rows

; Pointer to ROM character fonts
pTable8x8       DD      ?
pTable8x14      DD      ?
pTable8x16      DD      ?
pTable9x16      DD      ?

stRegs          TDPMIREGS       <>      ; DPMI registers

; Array of active windows
fpaWindow       DD      MAXHANDLE DUP(0)

;-----------------------------------------------------------
;
; Calls the video BIOS (int 10h)
;
; Input:
;       depends on call
; Output:
;       depends on call
; Note:
;       protected mode calls returns AX,BX,CX,DX,SI,DI,BP and ES
;
subVBios        PROC NEAR
        ASSUME  ds:NOTHING
        cmp     [boDPMI], TRUE
        je      @@DPMI
        int     10h                     ; Simple real-mode call
        jmp     @@Exit
@@DPMI:
        push    ds
        mov     ds, [wSegData]
        ASSUME  ds:TGX_TEXT

        mov     WORD PTR stRegs[rEAX], ax
        mov     WORD PTR stRegs[rEBX], bx
        mov     WORD PTR stRegs[rECX], cx
        mov     WORD PTR stRegs[rEDX], dx
        mov     WORD PTR stRegs[rESI], si
        mov     WORD PTR stRegs[rEDI], di
        mov     WORD PTR stRegs[rEBP], bp
        mov     WORD PTR stRegs[rES], es
        pop     ax
        mov     WORD PTR stRegs[rDS], ax
        mov     WORD PTR stRegs[rFS], ax
        mov     WORD PTR stRegs[rGS], ax
        pushf
        pop     ax
        mov     WORD PTR stRegs[rFLAGS], ax
        xor     ax, ax
        mov     WORD PTR stRegs[rXXX][0], ax
        mov     WORD PTR stRegs[rXXX][2], ax
        mov     WORD PTR stRegs[rSP], ax        ; Use DPMI default stack
        mov     WORD PTR stRegs[rSS], ax

        mov     ax, 0300h
        mov     bx, 0010h
        xor     cx, cx
        mov     es, [wSegData]
        mov     di, OFFSET stRegs
        int     31h

        mov     ds, [wSegData]
        mov     bx, WORD PTR stRegs[rES]
        verr    bx
        jz      @@Done
        mov     ax, 0002h
        int     31h
        mov     WORD PTR stRegs[rES], ax

@@Done:
        mov     ax, WORD PTR stRegs[rEAX]
        mov     bx, WORD PTR stRegs[rEBX]
        mov     cx, WORD PTR stRegs[rECX]
        mov     dx, WORD PTR stRegs[rEDX]
        mov     si, WORD PTR stRegs[rESI]
        mov     di, WORD PTR stRegs[rEDI]
        mov     bp, WORD PTR stRegs[rEBP]
        mov     es, WORD PTR stRegs[rES]

@@Exit:
        ret
subVBios        ENDP

;-----------------------------------------------------------
;
; Gets the pointer to a ROM font.
;
; Input:
;       BH      = font type
;       DS:DI   = pointer to doubleword buffer
; Output:
;       none (preserves BP)
;
subGetTable     PROC NEAR
        push    bp
        push    di
        mov     ax, 1130h
        call    subVBios
        pop     di
        mov     [di], bp
        mov     [di+2], es
        pop     bp
        ret
subGetTable     ENDP

;-----------------------------------------------------------
;
; Initializes the library.
;
; Input:
;       none
; Output:
;       AX = 0 on success, otherwise an error code
;
tgxInit         PROC PASCAL FAR
        LOCAL   Result:WORD,                    \
                Info:TVIDEOINFO
        USES    ds, si, es, di
        ASSUME  ds:NOTHING

; Check if already initialized
        mov     [Result], OK
        cmp     [boInit], TRUE
        je      @@Exit

; Check if running in protected mode under DPMI
        mov     ax, 1686h
        int     2Fh
        or      ax, ax
        jz      @@DPMI
        mov     ax, cs
        mov     ds, ax
         jmp    @@SetDS

; Create a data alias for CS and get selectors for segments A000h and B800h
@@DPMI:
        mov     [Result], ERR_DPMI
        mov     ax, 000Ah
        mov     bx, cs
        int     31h                             ; Create a data alias for CS
         jc     @@Exit
        mov     ds, ax
        mov     ax, 0002h
        mov     bx, 0A000h
        int     31h                             ; Get selector for seg. A000h
         jc     @@Exit
        mov     ds:[wSegA000h], ax
        mov     ax, 0002h
        mov     bx, 0B800h
        int     31h                             ; Get selector for seg. B800h
         jc     @@Exit
        mov     ds:[wSegB800h], ax
        mov     ds:[boDPMI], TRUE               ; Set DPMI flag

@@SetDS:
        ASSUME  ds:TGX_TEXT
        mov     [wSegData], ds

; Check if a VGA is present
        mov     [Result], ERR_VGA
        mov     ah, 1Bh
        xor     bx, bx
        push    ss
        pop     es
        lea     di, Info
; Get video info: this VBIOS call must work in protected mode!
        int     10h
        cmp     al, 1Bh
        jne     @@Exit

        lea     di, Info
        cmp     es:[di].bActiveDisplay, 07h     ; Check if a VGA is present
        jae     @@1
        cmp     es:[di].bAlternateDisplay, 07h
        jb      @@Exit
@@1:
        mov     [Result], ERR_VMODE
        cmp     es:[di].bVideoMode, 03h         ; Must be in color text mode
        ja      @@Exit

; Copy useful info in static variables
        mov     ax, es:[di].wScreenColumns
        mov     [wScreenWidth], ax
        mov     al, es:[di].bScreenRows
        xor     ah, ah
        mov     [wScreenHeight], ax
        mov     ax, es:[di].wCharacterHeight
        mov     [wCharHeight], ax

; Get pointers to ROM fonts
        mov     di, OFFSET pTable8x8
        mov     bh, 03h
        call    subGetTable
        mov     di, OFFSET pTable8x14
        mov     bh, 02h
        call    subGetTable
        mov     di, OFFSET pTable8x16
        mov     bh, 06h
        call    subGetTable
        mov     di, OFFSET pTable9x16
        mov     bh, 07h
        call    subGetTable

; Set init flag and return success
        mov     [boInit], TRUE
        mov     [Result], OK

@@Exit:
        mov     ax, [Result]
        ret
tgxInit         ENDP

;-----------------------------------------------------------
;
; Terminates the library.
;
; Input:
;       none
; Output:
;       none
;
tgxDone         PROC PASCAL FAR
        USES    ds, si, es, di
        ASSUME  ds:NOTHING

; Check if initializated
        cmp     [boInit], TRUE
        jne     @@Exit

        mov     ds, [wSegData]
        ASSUME  ds:TGX_TEXT

        mov     [boInit], FALSE
        mov     [boDPMI], FALSE

@@Exit:
        xor     ax, ax
        ret
tgxDone         ENDP

;-----------------------------------------------------------
;
; Gets a pixel.
;
; Input:
;       Window  = pointer to window
;       X, Y    = pixel coordinates
; Output:
;       AX      = 0 if pixel not set, else "or" mask to pixel
;
tgxGetPixel     PROC PASCAL FAR
        ARG     Window:DWORD,           \
                X:WORD,                 \
                Y:WORD
        USES    ds, si, es, di
        ASSUME  ds:NOTHING

        les     di, [Window]

        mov     ax, [Y]
        xor     dx, dx
        div     [wCharHeight]           ; DX = row in char, AX = column
        mov     bx, dx
        mov     cx, [X]
        mov     si, cx
        and     si, 07h
        shr     cx, 3                   ; CX = row
        mul     es:[di].wWidth
        add     ax, cx
        imul    ax, SIZE TCHAR
        add     di, ax
        mov     al, baOrPixelMask[si]
        and     al, es:[di+bx+SIZE TWINDOW+2]
        xor     ah, ah

        ret
tgxGetPixel     ENDP

;-----------------------------------------------------------
;
; Sets a pixel.
;
; Input:
;       Window  = pointer to window
;       X, Y    = pixel coordinates
; Output:
;       AX      = error code
;
tgxSetPixel     PROC PASCAL FAR
        ARG     Window:DWORD,           \
                X:WORD,                 \
                Y:WORD
        USES    ds, si, es, di
        ASSUME  ds:NOTHING

        les     di, [Window]

        mov     ax, [Y]
        xor     dx, dx
        div     [wCharHeight]           ; DX = row in char, AX = column
        mov     bx, dx
        mov     cx, [X]
        mov     si, cx
        and     si, 07h
        shr     cx, 3                   ; CX = row
        mul     es:[di].wWidth
        add     ax, cx
        imul    ax, SIZE TCHAR
        add     di, ax
        mov     al, baOrPixelMask[si]
        or      es:[di+bx+SIZE TWINDOW+2], al

        xor     ax, ax
        ret
tgxSetPixel     ENDP

;-----------------------------------------------------------
;
; Resets a pixel.
;
; Input:
;       Window  = pointer to window
;       X, Y    = pixel coordinates
; Output:
;       AX      = error code
;
tgxResetPixel   PROC PASCAL FAR
        ARG     Window:DWORD,           \
                X:WORD,                 \
                Y:WORD
        USES    ds, si, es, di
        ASSUME  ds:NOTHING

        les     di, [Window]

        mov     ax, [Y]
        xor     dx, dx
        div     [wCharHeight]           ; DX = row in char, AX = column
        mov     bx, dx
        mov     cx, [X]
        mov     si, cx
        and     si, 07h
        shr     cx, 3                   ; CX = row
        mul     es:[di].wWidth
        add     ax, cx
        imul    ax, SIZE TCHAR
        add     di, ax
        mov     al, baAndPixelMask[si]
        and     es:[di+bx+SIZE TWINDOW+2], al

        xor     ax, ax
        ret
tgxResetPixel   ENDP

;-----------------------------------------------------------
;
; Clears a window (does not update screen).
;
; Input:
;       Window  = pointer to window
; Output:
;       AX      = error code
;
tgxClear        PROC PASCAL FAR
        ARG     Window:DWORD
        USES    es, di
        ASSUME  ds:NOTHING

        les     di, [Window]            ; ES:DI -> window
        mov     bx, es:[di].wSize       ; Number of characters to clear
        add     di, SIZE TWINDOW        ; Skip header
        xor     ax, ax
        cld

@@Loop:
        inc     di
        inc     di                      ; Skip char header
        mov     cx, MAXCHARHEIGHT shr 1
        rep     stosw                   ; Clear char data
        dec     bx
        jnz     @@Loop

        xor     ax, ax
        ret
tgxClear        ENDP

;-----------------------------------------------------------
;
; Refreshes a window by updating the VGA font.
;
; Input:
;       Window  = pointer to window
; Output:
;       AX      = error code
;
tgxRefresh      PROC PASCAL FAR
        ARG     Window:DWORD
        USES    ds, si, es, di
        ASSUME  ds:NOTHING

        cmp     [boInit], TRUE
        jne     @@Exit

        lds     si, [Window]
        mov     bx, ds:[si].wSize       ; Number of characters to update
        add     si, SIZE TWINDOW        ; Skip header

        .xmem   on
        mov     es, [wSegA000h]         ; Enable hidden memory

        mov     dx, [wCharHeight]
        mov     ax, MAXCHARHEIGHT
        sub     ax, dx                  ; AX = extra bytes per character
        shr     dx, 1                   ; DX = character height in words

; Update font
        cld
@@Loop:
        mov     di, ds:[si]
        and     di, 00FFh
        shl     di, 5                   ; Character size is 32 bytes in VRAM
        mov     cx, dx
        inc     si
        inc     si
        rep     movsw
        add     si, ax                  ; Skip unused bytes
        dec     bx
        jnz     @@Loop

        .xmem   off

@@Exit:
        xor     ax, ax
        ret
tgxRefresh      ENDP

; INTERNAL USE ---------------------------------------------
;
; Clips a window.
;
; Input:
;       BX, AX  = X, Y coordinates of box
;       CX      = box width
;       DX      = box height
; Output:
;       CF      = set if box is entirely outside screen
;       BX, AX  = X, Y coordinates of box
;       CX      = clipped box width
;       DX      = clipped box height
;       SI      = number of bytes to skip before copying
;       DI      = extra bytes per buffer line
;
subClip PROC    NEAR
        ASSUME  ds:NOTHING
        push    bp
        xor     si, si
        xor     di, di

; Check height
        or      ax, ax
        jge     @@PositiveY
        add     dx, ax
         jl     @@Out
        neg     ax
        mov     bp, dx
        mul     cx
        mov     si, ax
        mov     dx, bp
        xor     ax, ax
@@PositiveY:
        cmp     ax, [wScreenHeight]
         jge    @@Out
        mov     bp, ax
        add     bp, dx
        sub     bp, [wScreenHeight]
         jl     @@DoneY
        sub     dx, bp
@@DoneY:

; Check width
        or      bx, bx
        jge     @@PositiveX
        add     cx, bx
         jl     @@Out
        neg     bx
        add     si, bx
        add     di, bx
        xor     bx, bx
@@PositiveX:
        cmp     bx, [wScreenWidth]
         jge    @@Out
        mov     bp, bx
        add     bp, cx
        sub     bp, [wScreenWidth]
         jl     @@DoneX
        sub     cx, bp
        add     di, bp
@@DoneX:

; Set return flag and exit
@@Done:
        imul    si, SIZE TCHAR
        imul    di, SIZE TCHAR
        clc
         jmp    @@Exit
@@Out:
        stc
@@Exit:
        pop     bp
        ret
subClip ENDP

;-----------------------------------------------------------
;
; Displays a window at a specified page.
;
; Input:
;       Window  = pointer to window
;       Page    = page to display window at
; Output:
;       AX      = error code
;
tgxShow         PROC PASCAL FAR
        ARG     Window:DWORD,                   \
                Page:BYTE:2
        LOCAL   WindowWidth:WORD,               \
                WindowHeight:WORD,              \
                WindowExtraBytes:WORD,          \
                ScreenExtraBytes:WORD
        USES    ds, si, es, di
        ASSUME  ds:NOTHING

        lds     si, [Window]

; Clip window
        mov     ax, ds:[si].iY
        mov     bx, ds:[si].iX
        mov     cx, ds:[si].wWidth
        mov     dx, ds:[si].wHeight
        push    si
        call    subClip
        mov     [WindowWidth], cx               ; Visible window width
        mov     [WindowHeight], dx              ; Visible window height
        mov     [WindowExtraBytes], di
        pop     cx
        jc      @@Exit                          ; Window not visible
        mul     [wScreenWidth]
        add     ax, bx
        shl     ax, 1
        mov     di, ax
        mov     es, [wSegB800h]                 ; ES:DI -> window position
        add     si, cx
        add     si, SIZE TWINDOW                ; DS:SI -> 1st char to display
        mov     ax, [wScreenWidth]
        sub     ax, [WindowWidth]
        shl     ax, 1
        mov     [ScreenExtraBytes], ax

; Write window to screen
        cld
@@LoopY:
        mov     cx, [WindowWidth]
@@LoopX:
        movsw
        add     si, SIZE TCHAR-2
        loop    @@LoopX
        add     di, [ScreenExtraBytes]
        add     si, [WindowExtraBytes]
        dec     [WindowHeight]
        jnz     @@LoopY

@@Exit:
        xor     ax, ax
        ret
tgxShow         ENDP

;-----------------------------------------------------------
;
; Grabs a character.
;
; Input:
;       Char    = character to grab
;       Dest    = pointer to TCHAR structure
; Output:
;       AX      = error code
;
tgxGrabChar     PROC PASCAL FAR
        ARG     Char:BYTE:2,            \
                Dest:DWORD
        USES    ds, si, es, di
        ASSUME  ds:NOTHING

        .xmem   on                              ; Enable access to fonts
        mov     ds, [wSegA000h]

        les     di, [Dest]
        mov     al, [Char]
        xor     ah, ah
        mov     es:[di], ax
        mov     si, ax
        shl     si, 5                           ; SI = char offset in video mem.
        inc     di
        inc     di                              ; Skip char header
        mov     cx, MAXCHARHEIGHT shr 1
        rep     movsw

        .xmem   off

@@Exit:
        xor     ax, ax
        ret
tgxGrabChar     ENDP

;-----------------------------------------------------------
;
; Copies font data from ROM to RAM.
;
; Input:
;       ES:DI   = pointer to character structure
; Output:
;       garbage in AX, BX, CX, DX
;
subGetROMFont   PROC NEAR
        ASSUME  ds:NOTHING
        push    ds
        push    si
        push    es
        push    di
        cld

        lds     si, [pTable8x8]         ; ROM 8x8 font
        cmp     [wCharHeight], 8
        je      @@GetROMFont
        lds     si, [pTable8x14]        ; ROM 8x14 font
        cmp     [wCharHeight], 14
        je      @@GetROMFont

; Height is 16 rows, check if special character
        mov     ah, es:[di]
        or      ah, ah
        jz      @@ROM8x16               ; Zero is not a special char
        lds     si, [pTable9x16]
@@Loop:
        lodsb
        cmp     al, ah                  ; Found?
        je      @@Done                  ; Yes, use this font
        add     si, [wCharHeight]       ; Go to next char
        or      al, al                  ; End of list?
        jnz     @@Loop                  ; No, continue

@@ROM8x16:
        lds     si, [pTable8x16]

; Get pointer to ROM font
@@GetROMFont:
        mov     al, es:[di]
        mul     BYTE PTR [wCharHeight]
        add     si, ax                  ; DS:SI -> character data

@@Done:
        inc     di
        inc     di                      ; Skip header
        mov     cx, [wCharHeight]
        cld
        rep     movsb

        pop     di
        pop     es
        pop     si
        pop     ds
        ret
subGetROMFont   ENDP

;-----------------------------------------------------------
;
; Initializes the various fields of a window structure.
;
; Input:
;       Window  = pointer to window
;       X, Y    = window coordinates
;       Width   = window width
;       Height  = window height
;       Attr    = attribute of all characters
;       CharList= pointer to character list (null terminated)
; Output:
;       AX      = error code
;
tgxInitWindow   PROC PASCAL FAR
        ARG     Window:DWORD,           \
                X:WORD,                 \
                Y:WORD,                 \
                Width:WORD,             \
                Height:WORD,            \
                Attr:BYTE:2,            \
                CharList:DWORD
        USES    ds, si, es, di
        ASSUME  ds:NOTHING

; Initialize header
        les     di, [Window]
        mov     ax, [X]
        mov     es:[di].iX, ax
        mov     ax, [Y]
        mov     es:[di].iY, ax
        mov     ax, [Width]
        mov     es:[di].wWidth, ax
        imul    ax, MAXCHARWIDTH
        mov     es:[di].wPixelWidth, ax
        mov     ax, [Height]
        mov     es:[di].wHeight, ax
        mul     [wCharHeight]
        mov     es:[di].wPixelHeight, ax
        mov     ax, di
        add     ax, SIZE TWINDOW
        mov     WORD PTR es:[di].fpChar[0], ax
        mov     WORD PTR es:[di].fpChar[2], es
        mov     ax, [Width]
        mul     [Height]
        mov     es:[di].wSize, ax
        mov     cx, ax

; Initialize character list
        mov     ah, [Attr]
        cld
        add     di, SIZE TWINDOW
        mov     bx, WORD PTR CharList[0]
        mov     dx, WORD PTR CharList[2]
        or      dx, dx
        jnz     @@String
        or      bx, bx
        jz      @@Exit
        jmp     @@CharLoop

@@String:
        lds     si, [CharList]
        mov     bl, 80h                 ; Default character
@@Loop:
        lodsb
        or      al, al
        jz      @@CharLoop
        mov     bl, al
        inc     bl                      ; Update default character
        stosw
        add     di, SIZE TCHAR - 2
        loop    @@Loop

@@CharLoop:
        mov     al, bl
        inc     bl
        stosw
        add     di, SIZE TCHAR - 2
        loop    @@CharLoop

@@Exit:
        xor     ax, ax
        ret
tgxInitWindow   ENDP

;-----------------------------------------------------------
;
; Restores ROM font for all characters in a window.
;
; Input:
;       Window  = pointer to window
; Output:
;       AX      = error code
;
tgxDoneWindow   PROC PASCAL FAR
        ARG     Window:DWORD
        USES    es, di
        ASSUME  ds:NOTHING

        les     di, [Window]
        mov     cx, es:[di].wSize
        add     di, SIZE TWINDOW

@@Loop:
        push    cx
        call    subGetROMFont
        pop     cx
        add     di, SIZE TCHAR
        loop    @@Loop

        les     di, [Window]
        push    es
        push    di
        call    tgxRefresh

@@Exit:
        xor     ax, ax
        ret
tgxDoneWindow   ENDP

;-----------------------------------------------------------
;
; Waits for vertical retrace start.
;
; Input:
;       none
; Output:
;       none
;
tgxWaitRetrace  PROC PASCAL FAR
        mov     dx, 3DAh
@@1:    in      al, dx
        test    al, 08h
        jnz     @@1
@@2:    in      al,dx
        test    al, 08h
        jz      @@2
        xor     ax, ax
        ret
tgxWaitRetrace  ENDP

TGX_TEXT        ENDS
END
